package com.sillycycle.bagleyd.threed;

/*
 * @(#)ThreedCanvas.java
 *
 * Copyright 1994 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import javax.swing.JFrame;

/**
 * The <code>ThreedCanvas</code> class does all the graphics for
 * <code>ThreedFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/threed.html
 */

public class ThreedCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	static final int MAX_SLICES = 16;
	static final int COLORS = 6;
	static final int DELTADISTANCE = 5;
	static final int DELTADEPTH = 5;
	static final int DELTADEGREES = 5;
	static final double EPSILON = 0.000001;
	static final double ASPECTRATIO = 1.0;
	static final int MAX_POLYGON_VERTICES = 10;

	Color       foreground, background;
	Color       borderColor;
	Color       currentForeground;
	Color[]     shadeColor = new Color[COLORS];
	transient ThreedPoint distance;
	boolean     mono, reverse, surface;
	int         object, numObjects;
	int         delay, numSlices;
	Point       coreSize;
	int         viewDistance;
	String      name;
	String[]    list;
	transient ThreedObject[] objects;
	transient ThreedAngle angle, deltaAngle;
	transient ThreedMath  math = new ThreedMath();
	Point       size;
	Point       center, currentPosition;

	JFrame      frame = null;
	boolean     firstPaint = false, resizePaint = false;
	boolean     framePaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false;
	String      stringSave = null;
	String      shadeNames = null;
	transient Image  bufferedImage = null;
	transient Graphics  bufferedGraphics = null;
	boolean     debug = false;

	static double radians(double x) {
		return Math.PI * x / ThreedInterface.ST_ANGLE;
	}

	static double degrees(double x) {
		return x / Math.PI * ThreedInterface.ST_ANGLE;
	}

	@Override
	protected String paramString() {
		return super.paramString() +
		",object=" + object +
		",surface=" + surface +
		",distance.x=" + distance.x +
		",distance.y=" + distance.y +
		",distance.z=" + distance.z +
		",angle.theta=" + angle.theta +
		",angle.phi=" + angle.phi +
		",angle.psi=" + angle.psi +
		",mono=" + mono +
		",reverse=" + reverse +
		",names=" + shadeNames +
		",shade0=" + shadeColor[0] +
		",shade1=" + shadeColor[1] +
		",shade2=" + shadeColor[2] +
		",shade3=" + shadeColor[3] +
		",shade4=" + shadeColor[4] +
		",shade5=" + shadeColor[5] +
		",borderColor=" + borderColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay;
	}

	void checkGraphicsThreeD() {
		int temp;

		if (distance.z < ThreedInterface.MINDISTANCE ||
				distance.z > ThreedInterface.MAX_DISTANCE) {
			temp = (ThreedInterface.MAX_DISTANCE +
				ThreedInterface.MINDISTANCE) / 2;
			System.out.println(distance.z +
				" distance out of bounds, use " +
				ThreedInterface.MINDISTANCE +
				".." +
				ThreedInterface.MAX_DISTANCE +
				", defaulting to " +
				temp);
			distance.z = temp;
		}
		if (angle.theta < ThreedInterface.MIN_DEGREES ||
				angle.theta >= ThreedInterface.NUM_DEGREES)
			angle.theta = (ThreedInterface.NUM_DEGREES +
				angle.theta) % ThreedInterface.NUM_DEGREES;
		if (angle.phi < ThreedInterface.MIN_DEGREES ||
				angle.phi >= ThreedInterface.NUM_DEGREES)
			angle.phi = (ThreedInterface.NUM_DEGREES +
				angle.phi) % ThreedInterface.NUM_DEGREES;
		if (object < 0 || object >= numObjects) {
			temp = -1;
			System.out.println(object +
				" object type number out of bounds, use 0.." +
				(numObjects - 1) +
				", defaulting to " +
				temp);
			object = temp;
		}
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside && currentForeground.equals(foreground))
				g.setColor(background.darker().darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside && currentForeground.equals(foreground))
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		if (g == null)
			return;
		draw3DFrame(show, 1, 1, coreSize.x - 2, coreSize.y - 2);
		if (show) {
			g.setColor(currentForeground);
		} else {
			g.setColor(background);
		}
		g.drawRect(0, 0, coreSize.x - 1, coreSize.y - 1);
	}

	void eraseGraphicsThreeD(Graphics g) {
		g.setColor(background);
		/* Do not erase frame */
		g.fillRect(3, 3, coreSize.x - 6, coreSize.y - 6);
	}

	void realizeObject(ThreedObject obj) {
		int vert, surf;
		double z;

		obj.rotateObject(deltaAngle);
		deltaAngle.theta = 0;
		deltaAngle.phi = 0;
		deltaAngle.psi = 0;
		for (vert = 0; vert < obj.vertices.length; vert++) {
			ThreedVertex vertex;
			ThreedPoint eye;
			Point screen;

			eye = new ThreedPoint(
				obj.local[vert].x + distance.x - obj.origin.x,
				obj.local[vert].y + distance.y - obj.origin.y,
				obj.local[vert].z + distance.z - obj.origin.z);
			/* Avoid divide by zero */
			if (eye.z < EPSILON && eye.z >= 0.0)
				eye.z = EPSILON;
			if (eye.z < 0.0 && eye.z > -EPSILON)
				eye.z = -EPSILON;
			screen = new Point((int) (center.x + eye.x *
				viewDistance / eye.z),
				(int) (center.y - ASPECTRATIO * eye.y *
				viewDistance / eye.z));
			vertex = new ThreedVertex(eye, screen);
			obj.vertex[vert] = vertex;
			if (debug) {
				System.out.println(vert + ": x " + eye.x +
					", y " + eye.y + ", z " + eye.z);
			}
		}
		for (surf = 0; surf < obj.surfaces.length; surf++) {
			ThreedSurface surfaceValue = obj.surfaces[surf];
			int mapIndex = surfaceValue.mapIndex;

			z = 0.0;
			for (vert = 0; vert < surfaceValue.numVertices;
					vert++, mapIndex++) {
				z += obj.vertex[obj.edges[mapIndex]].eye.z;
			}
			surfaceValue.averageDepth = z / surfaceValue.numVertices;
		}
	}

	void drawObject(Graphics g, ThreedObject obj) {
		int vert, vertex, mapIndex, clickSurface = -1;
		int[] pointsX = new int[MAX_POLYGON_VERTICES];
		int[] pointsY = new int[MAX_POLYGON_VERTICES];
		Point[] points = new Point[MAX_POLYGON_VERTICES + 1];
		double s1, s2, s3, sTotal;
		ThreedPoint v1, v2, v3;
		Color fillColor, lineColor;

		//dr = bufferObjects[0];
		/* compute position of object in world */
		if (debug) {
			System.out.println("drawObject " + object);
		}
		/* if (surface) */ {
			ThreedSurfaceSort.sort(obj.surfaces);
		}
		for (int surf = 0; surf < obj.surfaces.length; surf++) {
			mapIndex = obj.surfaces[surf].mapIndex;

			if (obj.surfaces[surf].clipped)
				continue;
			if (debug) {
				System.out.println("polygon " + surf);
				System.out.println("mapIndex " + mapIndex);
				System.out.println("e0 " + obj.edges[mapIndex]);
				System.out.println("e1 " + obj.edges[mapIndex + 1]);
				System.out.println("e2 " + obj.edges[mapIndex + 2]);
			}
			v1 = obj.vertex[obj.edges[mapIndex]].eye;
			v2 = obj.vertex[obj.edges[mapIndex + 1]].eye;
			v3 = obj.vertex[obj.edges[mapIndex + 2]].eye;
			s1 = -v1.x * (v2.y * v3.z - v3.y * v2.z);
			s2 = -v2.x * (v3.y * v1.z - v1.y * v3.z);
			s3 = -v3.x * (v1.y * v2.z - v2.y * v1.z);
			sTotal = s1+ s2 + s3;
			if (debug) {
				System.out.println(sTotal);
			}
			if (surface && sTotal < 0.0) {
				obj.surfaces[surf].visible = -1;
				continue;
			} else if (sTotal == 0.0) {
				obj.surfaces[surf].visible = 0;
			} else {
				obj.surfaces[surf].visible = 1;
			}
			for (vert = 0; vert < obj.surfaces[surf].numVertices;
					vert++, mapIndex++) {
				vertex = obj.edges[mapIndex];
				if (vert < MAX_POLYGON_VERTICES) {
					points[vert] = new Point(
						obj.vertex[vertex].screen.x,
						obj.vertex[vertex].screen.y);
				} else {
					System.out.println(
						"Ignoring extra polygon points");
					break;
				}
			}
			if (ThreedMath.pointInPolygon(currentPosition,
					points, vert)) {
				clickSurface = surf;
			}
		}
		for (int surf = 0; surf < obj.surfaces.length; surf++) {
			mapIndex = obj.surfaces[surf].mapIndex;
			if (obj.surfaces[surf].visible == -1 ||
					obj.surfaces[surf].clipped)
				continue;
			for (vert = 0; vert < obj.surfaces[surf].numVertices;
					vert++, mapIndex++) {
				vertex = obj.edges[mapIndex];
				if (vert < MAX_POLYGON_VERTICES) {
					/* Its not dynamic but we really after
					   speed here */
					pointsX[vert] = obj.vertex[vertex].screen.x;
					pointsY[vert] = obj.vertex[vertex].screen.y;
				} else {
					break;
				}
			}
			pointsX[vert] = pointsX[0];
			pointsY[vert] = pointsY[0];
			if (clickSurface >= 0 &&
					obj.surfaces[surf].depthIndex ==
					obj.surfaces[clickSurface].depthIndex) {
				fillColor = foreground;
				/*lineColor = foreground;*/
				lineColor = borderColor; 
			} else {
				fillColor = shadeColor[obj.surfaces[surf].color];
				/*lineColor = shadeColor[obj.surfaces[surf].color];*/
				lineColor = borderColor;
			}
			if (!surface || obj.surfaces[surf].visible == 0) {
				/*if (clickSurface < 0 ||
						obj.surfaces[surf].depthIndex !=
						obj.surfaces[clickSurface].depthIndex)*/ {
					g.setColor(fillColor);
					g.drawPolygon(pointsX, pointsY,
						vert);
				}
			} else {
				g.setColor(fillColor);
				g.fillPolygon(pointsX, pointsY, vert);
				g.setColor(lineColor);
				g.drawPolygon(pointsX, pointsY, vert + 1);
			}
		}
		for (int surf = 0; surf < obj.surfaces.length; surf++) {
			mapIndex = obj.surfaces[surf].mapIndex;
			if (obj.surfaces[surf].visible == -1 ||
					obj.surfaces[surf].clipped)
				continue;
			for (vert = 0; vert < obj.surfaces[surf].numVertices;
					vert++, mapIndex++) {
				vertex = obj.edges[mapIndex];
				if (vert < MAX_POLYGON_VERTICES) {
					/* Its not dynamic but we really after
					   speed here */
					pointsX[vert] = obj.vertex[vertex].screen.x;
					pointsY[vert] = obj.vertex[vertex].screen.y;
				} else {
					break;
				}
			}
			pointsX[vert] = pointsX[0];
			pointsY[vert] = pointsY[0];
			if (clickSurface >= 0 &&
					obj.surfaces[surf].depthIndex ==
					obj.surfaces[clickSurface].depthIndex) {
				lineColor = shadeColor[obj.surfaces[surf].color];
				g.setColor((surface && lineColor != null) ?
					lineColor : foreground);
				g.drawPolygon(pointsX, pointsY,
					vert);
			}
		}
	}

	void repaintGraphicsThreeD(Graphics g) {
		Graphics s = getGraphics();

		if (s == null)
			return;
		try {
			realizeObject(objects[object]);
			drawObject(g, objects[object]);
			/* cut out frame or it will flash */
			s.drawImage(bufferedImage,
				3, 3, coreSize.x - 3, coreSize.y - 3,
				3, 3, coreSize.x - 3, coreSize.y - 3,
				this);
			drawFrame(g, true);
		} finally {
			s.dispose();
		}
	}

	transient ThreedUniverseFallback threedUniverseFallback = null;
	void readThreeD() {
		ThreedUniverse universe = ThreedUniverse.readObjectsXML(ThreedInterface.XML_FILE, null);
		if (universe == null) {
			threedUniverseFallback = new ThreedUniverseFallback();
			universe = threedUniverseFallback.readStatic();
			System.err.println("Using fallback objects");
		}
		objects = universe.objects;
		numObjects = universe.objects.length;
		if (numObjects <= 0) {
			System.err.println("numObjects " + numObjects);
		}
		distance = universe.distance;
		angle = universe.angle;
		// post processing
		for (ThreedObject obj : objects) {
			for (int v = 0; v < obj.vertices.length; v++) {
				ThreedPoint vertex = obj.vertices[v];
				obj.local[v] = new ThreedPoint(vertex.x * obj.scale, vertex.y * obj.scale, vertex.z * obj.scale);
			}
		}
		if (debug) {
			checkThreeD();
		}
	}

	void checkThreeD() {
		final char DEFINE = '=';
		System.out.println("objects" + DEFINE + numObjects);
		System.out.println("distanceX" + DEFINE + distance.x);
		System.out.println("distanceY" + DEFINE + distance.y);
		System.out.println("distanceZ" + DEFINE + distance.z);
		System.out.println("angleTheta" + DEFINE + angle.theta);
		System.out.println("anglePhi" + DEFINE + angle.phi);
		System.out.println("anglePsi" + DEFINE + angle.psi);
		/*for (int i = 0; i < numObjects; i++) {
			System.out.println();
			System.out.println("NAME" + DEFINE + objectName);
			System.out.println("originX" + DEFINE + x);
			System.out.println("originY" + DEFINE + y);
			System.out.println("originZ" + DEFINE + z);
			System.out.println("Scale" + DEFINE + w);
			System.out.println("Convex" + DEFINE + n);
			System.out.println("name" + DEFINE +
				objects[i].name);
			System.out.println("numVertices" + DEFINE +
				objects[i].vertices.length);
			System.out.println("numSurfaces" + DEFINE +
				objects[i].surfaces.length);
			System.out.println("numEdges" + DEFINE +
				objects[i].numEdges);
			System.out.println();
			System.out.println("VERTICES" + DEFINE);
			for (int j = 0; j < objects[i].vertices.length; j++) {
				System.out.println(objects[i].local[j].x / w +
					"\t" + objects[i].local[j].y / w +
					"\t" + objects[i].local[j].z / w);
			}
			System.out.println();
			System.out.println("EDGES" + DEFINE);
			int l = 0;
			for (int j = 0; j < objects[i].surfaces.length; j++) {
				System.out.print(
					objects[i].surface[j].numVertices +
					"\t");
				for (int k = 0;
						k < objects[i].surface[j].numVertices;
						k++) {
					System.out.print(
						(objects[i].edges[l++] + 1) +
						"\t");
				}
				System.out.println();
			}
			System.out.println();
			System.out.println("SURFACES" + DEFINE);
			for (int j = 0; j < objects[i].surfaces.length; j++) {
				System.out.println(
					objects[i].surface[j].mapIndex +
					"\t" +
					objects[i].surface[j].depthIndex +
					"\t" +
					objects[i].surface[j].color);
			}
		}*/
	}

	void resizeThreeD() {
		coreSize = new Point(getSize().width, getSize().height);
		viewDistance = (coreSize.x + coreSize.y) / 2;
		center = new Point(coreSize.x / 2, coreSize.y / 2);
		currentPosition = new Point(coreSize.x * 2, coreSize.y * 2);
		bufferedImage = createImage(coreSize.x, coreSize.y);
		bufferedGraphics = bufferedImage.getGraphics();
	}

	void initializeThreeD() {
		int i;

		readThreeD();
		checkGraphicsThreeD();
		name = objects[object].name;
		list = new String[numObjects];
		for (i = 0; i < numObjects; i++)
			list[i] = objects[i].name;
		resizeThreeD();
	}

	void exposeThreeD() {
		eraseGraphicsThreeD(bufferedGraphics);
		repaintGraphicsThreeD(bufferedGraphics);
	}


	void selectThreeD(int x, int y) {
		currentPosition.x = x;
		currentPosition.y = y;
	}

	void motionThreeD(int x, int y) {
		int oldAngleTheta = angle.theta;
		int oldAnglePhi = angle.phi;

		if (currentPosition.x == 2 * coreSize.x &&
				currentPosition.y == 2 * coreSize.y)
			return;
		angle.theta += currentPosition.x - x;
		angle.phi += currentPosition.y - y;
		angle.theta = (64 * ThreedInterface.NUM_DEGREES + angle.theta) %
			ThreedInterface.NUM_DEGREES;
		angle.phi = (64 * ThreedInterface.NUM_DEGREES + angle.phi) %
			ThreedInterface.NUM_DEGREES;
		deltaAngle.theta = angle.theta - oldAngleTheta;
		deltaAngle.phi = angle.phi - oldAnglePhi;
		currentPosition.x = x;
		currentPosition.y = y;
		exposeThreeD();
		((ThreedFrame) frame).setThetaChanger(angle.theta);
		((ThreedFrame) frame).setPhiChanger(angle.phi);
	}

	void releaseThreeD() {
		currentPosition.x = 2 * coreSize.x;
		currentPosition.y = 2 * coreSize.y;
	}

	void surfaceThreeD() {
		if (mouseDown)
			return;
		((ThreedFrame) frame).callbackThreeD(
			ThreedInterface.ACTION_SURFACE);
	}

	void objectThreeD() {
		if (mouseDown)
			return;
		((ThreedFrame) frame).callbackThreeD(
			ThreedInterface.ACTION_OBJECT);
	}

	void setsurfaceThreeD(boolean surface3D) {
		surface = surface3D;
		exposeThreeD();
	}

	void setobjectThreeD(int object3D) {
		object = object3D;
		deltaAngle.theta = angle.theta - objects[object].orient.theta;
		deltaAngle.phi = angle.phi - objects[object].orient.phi;
		deltaAngle.psi = angle.psi - objects[object].orient.psi;
		exposeThreeD();
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
		numSlices = ((delay < MAX_SLICES) ? delay + 1 : MAX_SLICES);
	}

	public int getNumObjects() {
		return numObjects;
	}

	public void setSizeX(int distance) {
		this.distance.x = distance;
		exposeThreeD();
	}

	public void setSizeY(int distance) {
		this.distance.y = distance;
		exposeThreeD();
	}

	public void setSizeZ(int distance) {
		this.distance.z = distance;
		exposeThreeD();
	}

	public void setSizeTheta(int degree) {
		int oldAngle = angle.theta;

		this.angle.theta = degree;
		deltaAngle.theta = angle.theta - oldAngle;
		exposeThreeD();
	}

	public void setSizePhi(int degree) {
		int oldAngle = angle.phi;

		angle.phi = degree;
		deltaAngle.phi = angle.phi - oldAngle;
		exposeThreeD();
	}

	public void setSizePsi(int degree) {
		int oldAngle = angle.psi;

		angle.psi = degree;
		deltaAngle.psi = angle.psi - oldAngle;
		exposeThreeD();
	}

	void moveLowerPhiThreeD() {
		int oldAngle = angle.phi;

		angle.phi -= DELTADISTANCE;
		if (angle.phi < ThreedInterface.MIN_DEGREES)
			angle.phi += ThreedInterface.NUM_DEGREES;
		deltaAngle.phi = angle.phi - oldAngle;
		((ThreedFrame) frame).setPhiChanger(angle.phi);
		exposeThreeD();
	}

	void moveLowerThetaThreeD() {
		int oldAngle = angle.theta;

		angle.theta -= DELTADISTANCE;
		if (angle.theta < ThreedInterface.MIN_DEGREES)
			angle.theta += ThreedInterface.NUM_DEGREES;
		deltaAngle.theta = angle.theta - oldAngle;
		((ThreedFrame) frame).setThetaChanger(angle.theta);
		exposeThreeD();
	}

	void moveRaiseThetaThreeD() {
		int oldAngle = angle.theta;

		angle.theta += DELTADEGREES;
		if (angle.theta >= ThreedInterface.NUM_DEGREES)
			angle.theta -= ThreedInterface.NUM_DEGREES;
		deltaAngle.theta = angle.theta - oldAngle;
		((ThreedFrame) frame).setThetaChanger(angle.theta);
		exposeThreeD();
	}

	void moveRaisePhiThreeD() {
		int oldAngle = angle.phi;

		angle.phi += DELTADEGREES;
		if (angle.phi >= ThreedInterface.NUM_DEGREES)
			angle.phi -= ThreedInterface.NUM_DEGREES;
		deltaAngle.phi = angle.phi - oldAngle;
		((ThreedFrame) frame).setPhiChanger(angle.phi);
		exposeThreeD();
	}

	void moveLowerPsiThreeD() {
		int oldAngle = angle.psi;

		angle.psi -= DELTADISTANCE;
		if (angle.psi < ThreedInterface.MIN_DEGREES)
			angle.psi += ThreedInterface.NUM_DEGREES;
		deltaAngle.psi = angle.psi - oldAngle;
		((ThreedFrame) frame).setPsiChanger(angle.psi);
		exposeThreeD();
	}

	void moveRaisePsiThreeD() {
		int oldAngle = angle.psi;

		angle.psi += DELTADEGREES;
		if (angle.psi >= ThreedInterface.NUM_DEGREES)
			angle.psi -= ThreedInterface.NUM_DEGREES;
		deltaAngle.psi = angle.psi - oldAngle;
		((ThreedFrame) frame).setPsiChanger(angle.psi);
		exposeThreeD();
	}

	void moveRightThreeD() {
		double oldDistance = distance.x;

		distance.x += DELTADISTANCE;
		if (distance.x > ThreedInterface.MAX_DISTANCE) {
			distance.x = ThreedInterface.MAX_DISTANCE;
			if (Math.abs(distance.x - oldDistance) < EPSILON)
				return;
		}
		((ThreedFrame) frame).setXChanger((int) distance.x);
		exposeThreeD();
	}

	void moveLeftThreeD() {
		double oldDistance = distance.x;

		distance.x -= DELTADISTANCE;
		if (distance.x < ThreedInterface.MINDISTANCE) {
			distance.x = ThreedInterface.MINDISTANCE;
			if (Math.abs(distance.x - oldDistance) < EPSILON)
				return;
		}
		((ThreedFrame) frame).setXChanger((int) distance.x);
		exposeThreeD();
	}

	void moveUpThreeD() {
		double oldDistance = distance.y;

		distance.y += DELTADISTANCE;
		if (distance.y > ThreedInterface.MAX_DISTANCE) {
			distance.y = ThreedInterface.MAX_DISTANCE;
			if (Math.abs(distance.y - oldDistance) < EPSILON)
				return;
		}
		((ThreedFrame) frame).setYChanger((int) distance.y);
		exposeThreeD();
	}

	void moveDownThreeD() {
		double oldDistance = distance.y;

		distance.y -= DELTADISTANCE;
		if (distance.y < ThreedInterface.MINDISTANCE) {
			distance.y = ThreedInterface.MINDISTANCE;
			if (Math.abs(distance.y - oldDistance) < EPSILON)
				return;
		}
		((ThreedFrame) frame).setYChanger((int) distance.y);
		exposeThreeD();
	}

	void moveOutThreeD() {
		double oldDistance = distance.z;

		distance.z += DELTADEPTH;
		if (distance.z > ThreedInterface.MAX_DEPTH) {
			distance.z = ThreedInterface.MAX_DEPTH;
			if (Math.abs(distance.z - oldDistance) < EPSILON)
				return;
		}
		((ThreedFrame) frame).setZChanger((int) distance.z);
		exposeThreeD();
	}

	void moveInThreeD() {
		double oldDistance = distance.z;

		distance.z -= DELTADEPTH;
		if (distance.z < ThreedInterface.MIN_DEPTH) {
			distance.z = ThreedInterface.MIN_DEPTH;
			if (Math.abs(distance.z - oldDistance) < EPSILON)
				return;
		}
		((ThreedFrame) frame).setZChanger((int) distance.z);
		exposeThreeD();
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializeThreeD();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !selectPaint && !releasePaint) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizeThreeD();
			exposeThreeD();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(g, true);
			framePaint = false;
		}
		if (selectPaint) {
			selectPaint = false;
		}
		if (releasePaint) {
			releasePaint = false;
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}
